/* Miscellaneous utility predicates.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `misc'
**
**    This module contains some utility predicates used here and there.
**
**    Predicates defined in this module:
**
**    clair_intersection/3, clair_member/2, moreThanOnce/2,
**    setGlobal/1, subtraction/3, clair_union/3.
*/

/*
**    setGlobal(X)
**        - X is a term of the form f(Y).
**
**    The unit clause f(Y) is added to the database.  If a clause
**    f(Z) is already in the database, this is retracted first.
*/

setGlobal(X) :-
    functor(X, F, 1),
    Y =.. [F, _],
    (retract(Y) ; true),
    assertz(X),
    !.

/*
**    clair_member(E, L)
**
**    The usual list operations.
*/

clair_member(Elem, [Elem|_]) :- ! .
clair_member(Elem, [_|Tail]) :- clair_member(Elem, Tail).

/*
**    clair_intersection(X, Y, Z)
**    clair_union(X, Y, Z)
**    subtraction(X, Y, Z)
**        - X, Y, Z are lists representing sets.
**
**    Implementation of the corresponding set operations.
**    For example:
**
**        clair_union(A, B, C) <=> C represents the union of the sets
**                                 represented by A and B.
*/

clair_union([X|R], Y, Z) :-
    clair_member(X, Y),
    !,
    clair_union(R, Y, Z).
clair_union([X|R], Y, [X|Z]) :-
    clair_union(R, Y, Z).
clair_union([], X, X).

clair_intersection([Head|Tail], List, [Head|Remainder]) :-
    clair_member(Head, List),
    !,
    clair_intersection(Tail, List, Remainder).
clair_intersection([_|Tail], List, Res) :-
    clair_intersection(Tail, List, Res).
clair_intersection([], _, []).

subtraction([X|R], Y, Z) :-
    clair_member(X, Y),
    !,
    subtraction(R, Y, Z).
subtraction([X|R], Y, [X|Z]) :-
    subtraction(R, Y, Z).
subtraction([], _, []).

/*
**    moreThenOnce(L, D)
**        - L is a list;
**        - D is the list of elements that occur in D more than once.
*/

moreThanOnce([H|T], D) :-
    clair_member(H, T), !,
    moreThanOnce(T, C),
    clair_union([H], C, D).
moreThanOnce([_|T], D) :-
    moreThanOnce(T, D).
moreThanOnce([], []).

/*
**    existingFile(X)
**       - X is an atom.
**
**    Succeeds if a file named X exists in the current working directory.
*/

existingFile(X) :-
    catch(see(X), _, fail),
    seen.

/*
**    writeCpuTime(X)
**       - X is a non-negative integer expressing milliseconds of CPU time.
**
**    Writes the number of seconds corresponding to X.
*/

writeCpuTime(X) :-
    Seconds is X // 1000,
    Hundredth is ((X mod 1000)+5) // 10,
    write(Seconds),
    write('.'),
    (Hundredth < 10 ->
	write(0)
    ;
	true
    ),
    write(Hundredth).
