/* The tokenizer for CLAIR's SIL and SFL languages.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `parse'
**
**    This module contains the parser for the SIL and SFL languages.
**    It is based on DCGs (Definite Clause Grammars).
**
**    The parser accepts a list of tokens produced by the `tokenize'
**    module and produces an abstract syntax tree of the program
**    represented as a Prolog term.
**
**    Parsing is done top-down /with/ backtracking for the only
**    reason to allow the omission of the else branch in the
**    conditional command.  If we decide to eliminate this possibility
**    the grammar will become LL(1) and the parsing could thus be
**    deterministic (predictive).
**
**    Predicates defined in this module:
**
**    afterBegin/3, afterSide/5, alphaNumeric/1, boolConstant/3,
**    compStat/3, decl/3, decl1/3, decl2/3, decl3/3, expr/3,
**    exprList/3, factor/3, factorAfterId/4, forDirection/5,
**    formList/3, formal/3, funcHead/5, id/3, init/3, lexId/1,
**    limit/3, limits/3, maybeElse/3, maybeInit/4, moreDecl/4,
**    moreDecl1/4, moreDecl2/4, moreExpr/4, moreLimits/4, moreNeEL/4,
**    moreNeFL/4, morePrint/3, moreSExpr/4, moreSList/4, moreTerm/4,
**    neExprList/3, neFormList/3, num/3, optionalStat/3, procHead/4,
**    program/4, reserved/1, simpleExpr/3, simpleType/3, stat/3,
**    statAfterId/4, statList/3, term/3, type/3, unsignedNum/3,
**    varAfterId/4. 
*/

:- discontiguous(decl3/3).
:- discontiguous(factor/3).
:- discontiguous(stat/3).

/*
**    Simple types.
**    -------------
*/

simpleType(int) --> [integer], !.
simpleType(bool) --> [boolean], !.

/*
**    Types, including arrays.
**    ------------------------
*/

type(T) --> simpleType(T), !.

type(array(Limits, T)) --> {language(imperative)},
    [array], !, ['['], limits(Limits), [']'], [of], simpleType(T).

limit(b(L, H)) --> num(L), ['..'], num(H).

limits(X) --> limit(Y), moreLimits([Y], X).

moreLimits(Y, X) --> [','], !, limit(Z), {append(Y, [Z], W)}, moreLimits(W, X).
moreLimits(X, X) --> [].

/*
**    Identifiers.
**    ------------
*/

id(X) --> [X], {atom(X), lexId(X), !, \+ reserved(X)}.

lexId(X) :- name(X, [C|L]),            /* The first character can be:      */
               ((C >= 97, C =< 122) ;  /*   a lower-case letter or         */
                (C >= 65, C =< 90)  ;  /*   an upper-case letter or        */
                 C = 95),              /*   the underscore character `_'.  */
                   alphaNumeric(L).

alphaNumeric([C|T]) :- ((C >= 97, C =< 122) ;
                 (C >= 65, C =< 90)  ;
                 (C >= 48, C =< 57)  ;
                  C = 95),
                    alphaNumeric(T).
alphaNumeric([]).

/*
**    Variable or array reference (after id).
**    ---------------------------------------
*/

varAfterId(X, aref(X, El)) --> {language(imperative)},
    ['['], !, neExprList(El), [']'].

varAfterId(X, i(X)) --> [].

/*
**    Numerical constants.
**    --------------------
*/

unsignedNum(X) --> [X], {integer(X)}, !.

num(X) --> unsignedNum(X), !.
num(X) --> [+], !, unsignedNum(X).
num(X) --> [-], !, unsignedNum(Y), {X is -Y}.

/*
**    Boolean constants.
**    ------------------
*/

boolConstant(tt) --> [true], !.
boolConstant(ff) --> [false], !.

/*
**    Expressions.
**    ------------
*/

expr(X) --> simpleExpr(Y), moreExpr(Y, X).

/*
**    moreExpr(Y, X)
**        - Y is the tree built so far in this sequence
**            of terms (inherited);
**        - X is the final tree (synthesized).
**
**    The same technique (that also allows to realize
**    left-associativity of the operators) is used several
**    times in the sequel.
**    See Aho, Sethi, Ullman:
**         "Compilers: principles, techniques and tools", Chapter 5.
*/

moreExpr(Y, X) --> [==], !, simpleExpr(Z), moreExpr(eq(Y, Z), X).
moreExpr(Y, X) --> [<>], !, simpleExpr(Z), moreExpr(ne(Y, Z), X).
moreExpr(Y, X) --> [<],  !, simpleExpr(Z), moreExpr(lt(Y, Z), X).
moreExpr(Y, X) --> [<=], !, simpleExpr(Z), moreExpr(le(Y, Z), X).
moreExpr(Y, X) --> [>=], !, simpleExpr(Z), moreExpr(ge(Y, Z), X).
moreExpr(Y, X) --> [>],  !, simpleExpr(Z), moreExpr(gt(Y, Z), X).
moreExpr(X, X) --> [].

simpleExpr(X) --> term(Y), moreSExpr(Y, X).

moreSExpr(Y, X) --> [+],  !, term(Z), moreSExpr(plus(Y, Z), X).
moreSExpr(Y, X) --> [-],  !, term(Z), moreSExpr(minus(Y, Z), X).
moreSExpr(Y, X) --> [or], !, term(Z), moreSExpr(or(Y, Z), X).
moreSExpr(X, X) --> [].

term(X) --> factor(Y), moreTerm(Y, X).

moreTerm(Y, X) --> [*],   !, factor(Z), moreTerm(prod(Y, Z), X).
moreTerm(Y, X) --> [div], !, factor(Z), moreTerm(div(Y, Z), X).
moreTerm(Y, X) --> [mod], !, factor(Z), moreTerm(mod(Y, Z), X).
moreTerm(Y, X) --> [and], !, factor(Z), moreTerm(and(Y, Z), X).
moreTerm(X, X) --> [].

factor(X) --> unsignedNum(X), !.
factor(X) --> boolConstant(X), !.
factor(X) --> ['('], !, expr(X), [')'].

factor(X) --> id(Y), !, factorAfterId(Y, X).

/*
**    Function call.
**    --------------
*/

factorAfterId(Y, funcall(Y, ExprList)) -->
    ['('], !, exprList(ExprList), [')'].

/*
**    Identifier or array reference.
**    ------------------------------
*/

factorAfterId(Y, X) --> varAfterId(Y, X), !.

/*
**    Unary operators.
**    ----------------
*/

factor(not(X)) -->    [not], !, factor(X).
factor(X) -->         [+],   !, factor(X).
factor(uminus(X)) --> [-],   !, factor(X).

/*
**    Expression block.
**    -----------------
*/

factor(let(D, E)) --> {language(functional)}, [let], !, decl(D), [in], expr(E).

/*
**    Conditional expression.
**    -----------------------
*/

factor(if(E, E1, E2)) --> [if], !, expr(E), [then], expr(E1), [else], expr(E2).

/*
**    Expressions with side-effects.
**    ------------------------------
*/

factor(side(D, C, E)) --> {language(imperative)},
    [expr], !, afterSide(D, C, E).

afterSide(D, C, E)   --> decl(D), !, [;], optionalStat(C), [result], expr(E).
afterSide(nil, C, E) --> optionalStat(C), !, [result], expr(E).

/*
**    User input.
**    -----------
*/

factor(input(X)) --> [input], !, simpleType(X).

/*
**    List of (actual) expressions.
**    -----------------------------
*/

exprList(X) --> neExprList(X).
exprList([]) --> [].

neExprList(X) --> expr(Y), moreNeEL([Y], X).

moreNeEL(Y, X) --> [','], !, expr(Z), {append(Y, [Z], W)}, moreNeEL(W, X).
moreNeEL(X, X) --> [].

/*
**    Declarations.
**    -------------
*/

decl(X) --> decl1(Y), moreDecl(Y, X).

/*
**    Sequential composition.
**    -----------------------
*/

moreDecl(Y, X) --> [;], decl1(Z), moreDecl(seq(Y, Z), X).
moreDecl(X, X) --> [].

decl1(X) --> decl2(Y), moreDecl1(Y, X).

/*
**    Parallel composition.
**    ---------------------
*/

moreDecl1(Y, X) --> ['|'], !, decl2(Z), moreDecl1(parallel(Y, Z), X).
moreDecl1(X, X) --> [].

decl2(X) --> decl3(Y), moreDecl2(Y, X).

/*
**    Private composition.
**    --------------------
*/

moreDecl2(Y, X) --> [->], !, decl3(Z), moreDecl2(private(Y, Z), X).
moreDecl2(X, X) --> [].

/*
**    Simple declaration.
**    -------------------
*/

decl3(def(X, T, E)) --> {language(functional)},
                          id(X), !, [:], simpleType(T), init(E).

/*
**    Constant declaration.
**    ---------------------
*/

decl3(con(X, T, E)) --> {language(imperative)},
                          [const], !, id(X), [:], simpleType(T),
                               init(E).

/*
**    Variable declaration.
**    ---------------------
*/

decl3(var(X, T, E)) --> {language(imperative)},
                          [var], !, id(X), [:], type(T),
                              maybeInit(T, E).

/*
**    Parentheses around declarations.
**    --------------------------------
*/

decl3(X) --> ['('], !, decl(X), [')'].

/*
**    Recursive declaration.
**    ----------------------
*/

decl3(rec(X)) --> [rec], !, decl(X).

/*
**    Variable initialization (both optional and mandated).
**    -----------------------------------------------------
*/

init(E) --> [=], expr(E).

maybeInit(_, E) --> init(E).
maybeInit(T, X) --> [],
  {
    (
      (T=array(_, TypeArray), RealType=TypeArray)
    ;
      RealType=T
    ),
    (
      (RealType=int, X=0)
    ;
      (RealType=bool, X=ff)
    )
  }.

/*
**    Function declaration.
**    ---------------------
*/

decl3(fundef(F, Formals, T, E)) -->
    funcHead(F, Formals, T), !, [=], expr(E).

/*
**    Procedure declaration.
**    ----------------------
*/

decl3(procdef(P, Formals, C)) --> {language(imperative)},
    procHead(P, Formals), !, compStat(C).

/*
**    Heading of procedure declaration.
**    ---------------------------------
*/

procHead(P, Formals) --> [procedure],
    id(P), ['('], formList(Formals), [')'].

/*
**    Heading of function declaration.
**    --------------------------------
*/

funcHead(F, Formals, T) --> [function],
    id(F), ['('], formList(Formals), [')'], [:], simpleType(T).

/*
**    Formal parameters.
**    ------------------
*/

formList(X) --> neFormList(X).
formList([]) --> [].

neFormList(X) --> formal(Y), moreNeFL([Y], X).

moreNeFL(Y, X) --> [','], !, formal(Z), {append(Y, [Z], W)}, moreNeFL(W, X).
moreNeFL(X, X) --> [].

formal(ival(X, T)) --> {language(imperative)},    /* call by value        */
    id(X), !, [:], simpleType(T).
formal(fval(X, T)) --> {language(functional)},    /* call by value        */ 
    id(X), !, [:], simpleType(T).
formal(name(X, T)) -->                            /* call by name         */
    [name], !, id(X), [:], simpleType(T).
formal(ref(X, T)) --> {language(imperative)},     /* call by reference    */
    [ref], !, id(X), [:], simpleType(T).
formal(const(X, T)) --> {language(imperative)},   /* call by constant     */
    [const], !, id(X), [:], simpleType(T).
formal(copy(X, T)) --> {language(imperative)},    /* call by value-result */
    [copy], !, id(X), [:], simpleType(T).
formal(funcpar(F, Formals, T)) -->                /* functional parameter */ 
    funcHead(F, Formals, T), !.
formal(procpar(P, Formals)) -->                   /* procedural parameter */
    {language(imperative)}, 
    procHead(P, Formals), !.

/*
**    Commands.
**    ---------
*/

/*
**    Command block or composite command.
**    -----------------------------------
*/

compStat(X) --> [begin], afterBegin(X).

afterBegin(block(D, C)) --> decl(D), !, [;], optionalStat(C), [end].
afterBegin(C) --> optionalStat(C), !, [end].

/*
**    Optional command sequence.
**    --------------------------
*/

optionalStat(X) --> statList(X).
optionalStat(nop) --> [].

/*
**    Command sequence.
**    -----------------
*/

statList(X) --> stat(Y), moreSList(Y, X).

moreSList(Y, X) --> [;], stat(Z), moreSList(seq(Y, Z), X).
moreSList(X, X) --> [].

/*
**    Commands.
**    ---------
*/

/*
**    Procedure call or assignment.
**    -----------------------------
*/

stat(X) -->
    id(Y), !, statAfterId(Y, X).

/*
**    Procedure call.
**    ---------------
*/

statAfterId(Y, proccall(Y, ExprList)) -->
    ['('], !, exprList(ExprList), [')'].

/*
**    Assignment.
**    -----------
*/

statAfterId(Y, ass(X, E)) -->
    varAfterId(Y, X), !, [:=], expr(E).

/*
**    Composite command.
**    ------------------
*/

stat(X) -->
    compStat(X), !.

/*
**    Conditional command.
**    ---------------------
*/

stat(if(E, C1, C2)) -->
    [if], expr(E),
    [then], stat(C1),
    maybeElse(C2).

maybeElse(C) --> [else], !, stat(C).
maybeElse(nop) --> [].

/*
**    While command.
**    --------------
*/

stat(while(E, C)) -->
    [while], !, expr(E), [do], stat(C).

/*
**    Print command.
**    --------------
*/

stat(print(Format, El)) -->
    [print], !, ['('], [Format], morePrint(El).
morePrint(El) -->
    [','], !, exprList(El), [')'].
morePrint([]) -->
    [')'].

/*
**    For command.
**    ------------
**
**    N.B.: The `for' command does /not/ exist in the abstract syntax.
**          Instead, the parser expands it as a `while' cycle as follows:
**
**                for X := E1 to (downto) E2 do C
**
**          is translated to
**
**                X := E1;
**                begin
**                  const $limit : integer = E2;
**                  while X <= (>=) $limit do begin
**                    begin
**                      const X : integer = X;
**                      C
**                    end;
**                    X := X + (-) 1
**                  end
**                end
**
**          Notice that C (the for's body) is executed in an environment
**          where X (the for's index) is a constant, and thus cannot be
**          modified.  The for command of SLI is thus syntactically
**          end semantically identical to Pascal's for.
*/

stat(seq(ass(i(X), E1),
            block(con('$limit', int, E2),
              while(Comparison,
                seq(block(con(X, int, i(X)), C), ass(i(X), Update))))))
  --> [for], !, id(X), [:=], expr(E1), forDirection(X, Comparison, Update),
          expr(E2), [do], stat(C).

forDirection(X, le(i(X), i('$limit')), plus(i(X), 1)) -->  [to], !.
forDirection(X, ge(i(X), i('$limit')), minus(i(X), 1)) --> [downto], !.

/*
**    Program (initial syntactic categoryc for SLI).
**    ----------------------------------------------
*/

program(Name, Command) --> [program], id(Name), compStat(Command).

/*
**    Reserved words.
**    ---------------
*/

reserved(program) :- language(imperative).
reserved(const) :- language(imperative).
reserved(var) :- language(imperative).
reserved(array) :- language(imperative).
reserved(of) :- language(imperative).
reserved(begin) :- language(imperative).
reserved(end) :- language(imperative).
reserved(expr) :- language(imperative).
reserved(result) :- language(imperative).
reserved(ref) :- language(imperative).
reserved(copy) :- language(imperative).
reserved(while) :- language(imperative).
reserved(do) :- language(imperative).
reserved(for) :- language(imperative).
reserved(to) :- language(imperative).
reserved(downto) :- language(imperative).
reserved(procedure) :- language(imperative).
reserved(print) :- language(imperative).

reserved(let) :- language(functional).
reserved(in) :- language(functional).

reserved(integer).
reserved(boolean).
reserved(input).
reserved(name).
reserved(rec).
reserved(and).
reserved(or).
reserved(mod).
reserved(div).
reserved(if).
reserved(then).
reserved(else).
reserved(true).
reserved(false).
reserved(not).
reserved(function).
