/* Predicates for run-time support.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `runtime'
**
**    This module contains the predicates implementing run-time
**    support for the SIL and SFL languages.
**
**    Predicates defined in this module:
**
**    arrayIndex/3, atomChar/1, bomb/0, changeStore/4, delete/3,
**    digit/2, dispose/3, disposeArray/4, doInput/2, doOutput/1,
**    doPrint/2, doPrint1/2, getNewLoc/1, getStore/3, makeArray/4,
**    makeEnv/3, makeEnvStore/4, makeRArray/3, makeRecEnv/3,
**    numElem/2, readAtom/1, readAtom_1/2, readInteger/1, readInteger_2/2,
**    readInteger_3/3, readString/1, readString_1/2, returnAtom/2,
**    semAnd/3, semOr/3. 
*/

/*
**    Run-time support for `input' expressions.
**    -----------------------------------------
**
**    doInput(T, K)
**        - T is instantiated to `int' or `bool';
**        - K is a free variable.
**
**    A constant is read from the current input stream and K
**    is instantiated to the corresponding representation.
**
**    If T=int an integer is required and K is instantiated
**    accordingly: for instance, if the user digits '1' '2' <CR>
**    then K is instantiated to 12.
**
**    If T=bool a boolean value, `true' or `false', is required and
**    K is instantiated to `tt' or `ff', respectively.
**
**    If the input does not correspond to the type, doInput/2 invokes
**    itself recursively until something legal is entered.
*/
 
doInput(int, N) :- prompt(Old, 'Give me an integer> '),
    readInteger(N), prompt(_, Old).

doInput(bool, B) :- prompt(Old, 'Give me a boolean> '),
    readAtom(A),
    ((((A=true, !, B=tt); (A=false, !, B=ff)), prompt(_, Old));
        (write('Sorry, a boolean is expected.'), nl, doInput(bool, B))).

/*
**    Run-time support for the `print' command.
**    -----------------------------------------
**
**    doPrint(F, El)
**        - F  is a string (a list of ASCII codes);
**        - El is a (possibly empty) list of constants.
**
**    F is a string containin two kinds of objects:
**    normal characters, that are simply printed as they are,
**    and output specifications.
**    Output specifications are introduced by the character '#'.
**    After '#' we may have:
**
**        1) the character 'n':
**             a new line is initiated;
**
**        2) the character 'e':
**             a constant in El is printed;
**
**        3) the character '#':
**             the character '#' s printed;
**
**        4) any other character C:
**             the characters '#' and C are printed.
*/

doPrint(FormAtom, Cl) :-
    name(FormAtom, Format),
    doPrint1(Format, Cl).

doPrint1([35, 35 |Rest], Cl) :-
    put_code(35),
    doPrint1(Rest, Cl).
doPrint1([35, 110|Rest], Cl) :- !,
    nl,
    doPrint1(Rest, Cl).
doPrint1([35, 101|Rest], [C|Cl]) :- !,
    doOutput(C),
    doPrint1(Rest, Cl).
doPrint1([35, 101|Rest], []) :- !,
    write('*no value*'),
    doPrint1(Rest, []).
doPrint1([Char|Chars], Cl) :- !,
    put_code(Char),
    doPrint1(Chars, Cl).
doPrint1([], _).

/*
**    doOutput(X)
**        - X is a constant (tt, ff or an integer).
**
**    Prints `true' if X=tt, `false' if X=ff, and X otherwise
**    (X is assumed to be instantiated to an integer in this case).
*/

doOutput(tt) :-
    write(true).
doOutput(ff) :-
    write(false).
doOutput(X) :-
    write(X).

/*
**    Run-time support for boolean operators.
**    ---------------------------------------
*/

semOr(ff, ff, ff) :- !.
semOr(_, _, tt).

semAnd(tt, tt, tt) :- !.
semAnd(_, _, ff).

/*
**    Run-time support for arrays.
**    ----------------------------
*/

/*
**    numElem(Limits, N)
**       - Limits is a list of array limits ([b(L1, H1), ...]);
**       - N is an integer.
**
**    True if N is the number of elements of an array with limits
**    given by Limits.
**    A simple recurrence relation is used for the computation.
*/

numElem([b(L, H)|Tail], N) :-
    numElem(Tail, M),
    N is (H-L+1)*M.
numElem([], 1).

/*
**    makeArray(Limits, Base, Init, Store)
**       - Limits is a list of array limits ([b(L1, H1), ...]);
**       - Base   is an integer;
**       - Init   is a constant (tt, ff or an integer);
**       - Store  is a store.
**
**    Given Limits and Init, the remaining variables are instantiated
**    so that Store is a store containing as many consecutive
**    locations (*) as required to memorize an array with limits given
**    by Limits (see numElem/2), all containing the constant Init, the
**    first of which is Base.
**
**    (*) Consecutivity of the locations is a property of getNewLoc/1.
*/

makeArray(Limits, Base, Init, [s(Base, Init)|Rest]) :-
    getNewLoc(Base), numElem(Limits, N), makeRArray(N, Init, Rest).

makeRArray(1, _, []).
makeRArray(N, Init, [s(L, Init)|Rest]) :-
    getNewLoc(L), M is N-1, makeRArray(M, Init, Rest).

/*
**    arrayIndex(Limits, Indexes, Offset)
**       - Limits  is a list of array limits ([b(L1, H1), ...]);
**       - Indexes is a list of integer indices;
**       - Offset  is an integer.
**
**    Computes the offset of the cell of an array with limits given by Limits,
**    and indexed by Indexes. It uses the following recurrence:
**
**    let L_1..H_1, ..., L_m..H_m the limits of the array;
**        I_1, ..., I_m           the list of indices;
**        N_i = H_i-L_i+1         for each i = 1, ..., m;
**
**    then Offset = E_m, where
**
**        E_0 = 0
**        E_k = E_{k-1}*N_k + I_k-L_k.
*/

arrayIndex([b(L, H)|T1], [I|T2], Offset) :-
    ((I >= L, I =< H) ->
        (arrayIndex(T1, T2, R), Offset is R*(H-L+1)+I-L)
    ;
        (bomb, !, fail)).

arrayIndex([], [], 0).

/*
**    bomb
**
**    Invoked by arrayIndex/3 when an index in an array reference
**    exceeds the array limits.
**    Yeah, it is a little brutal...
*/

bomb :-
    nl, write('Panic: array index out of bounds!!!'), nl, abort.

/*
**    Run-time support for recursive declarations.
**    --------------------------------------------
**
**    makeRecEnv(X, Y, Z) -
**        - X is an environment to be made recursive;
**        - Y is the environment yet to be transformed;
**        - Z is the resulting environments.
**
**    Implements the final phase of the processing of recursive
**    declarations (see the documentation for details).
*/

/*
**    Functions.
*/

makeRecEnv(E,
	   [e(X, func(Formals, Body)) | T1],
	   [e(X, func(Formals, let(rec(E1), Body))) | T2]) :-
    formDV(Formals, V),
    mask(E, V, E1),
    makeRecEnv(E, T1, T2).

/*
**    Procedures.
*/

makeRecEnv(E,
	   [e(X, proc(Formals, Body)) | T1],
	   [e(X, proc(Formals, block(rec(E1), Body))) | T2]) :-
    formDV(Formals, V),
    mask(E, V, E1),
    makeRecEnv(E, T1, T2).

/*
**    Other denotable values.
*/

makeRecEnv(E, [H | T1], [H | T2]) :-
    makeRecEnv(E, T1, T2).

makeRecEnv(_, [], []).

/*
**    Run-time support for SFL's parameter passing.
**    ---------------------------------------------
**
**    makeEnv(Form, Act, Env)
**        - Form is a list of formal parameters;
**        - Act  is a list of already processed actual parameters;
**        - Env  is an environment.
**
**    Implements the final phase of SFL's parameter passing.
**    True if Env is the environment obtained by associating each
**    formal parameter in Form to the corresponding actual parameter
**    in Act.
*/

/*
**    Call by value.
*/

makeEnv([fval(X, _) | Form], [C | Ac], [e(X, C) | Env]) :-
    makeEnv(Form, Ac, Env).

/*
**    Call by name.
*/

makeEnv([name(X, _) | Form], [C | Ac], [e(X, C) | Env]) :-
    makeEnv(Form, Ac, Env).

/*
**    Functional parameters.
*/

makeEnv([funcpar(F, _, _) | Form], [Func | Ac],  [e(F, Func) | Env]) :-
    makeEnv(Form, Ac, Env).

makeEnv([], [], []).

/*
**    Run-time support for SIL's parameter passing.
**    ---------------------------------------------
**
**    makeEnvStore(Form, Act, Env, Store)
**        - Form  is a list of formal parameters;
**        - Act   is a list of already processed actual parameters;
**        - Env   is an environment;
**        - Store is a store.
**
**    Implements the final phase of SIL's parameter passing.
**    True if Env and Store are, respectively, the environment and the
**    store resulting from parameter passing of Act to Form (see the
**    documentation for details).
*/

/*
**    Call by value.
*/

makeEnvStore([ival(X, T) | Form], [C | Ac],
             [e(X, d(T, l(L))) | Env],  [s(L, C) | Store]) :-
    getNewLoc(L), makeEnvStore(Form, Ac, Env, Store).

/*
**    Call by name.
*/

makeEnvStore([name(X, T) | Form], [C | Ac],
             [e(X, d(T, C)) | Env], Store) :-
    makeEnvStore(Form, Ac, Env, Store).

/*
**    Call by reference.
*/

makeEnvStore([ref(X, T) | Form], [L | Ac],
             [e(X, d(T, r(L))) | Env], Store) :-
    makeEnvStore(Form, Ac, Env, Store).

/*
**    Call by constant.
*/

makeEnvStore([const(X, T) | Form], [C | Ac],
             [e(X, d(T, C)) | Env], Store) :-
    makeEnvStore(Form, Ac, Env, Store).

/*
**    Call by value-result (copy in - copy out).
*/

makeEnvStore([copy(X, T) | Form], [lr(L1, C) | Ac],
             [e(X, d(T, c(L1, L2))) | Env], [s(L2, C) | Store]) :-
    getNewLoc(L2),
    makeEnvStore(Form, Ac, Env, Store).

/*
**    Functional parameters.
*/

makeEnvStore([funcpar(F, FForm, T) | Form], [Func | Ac],
             [e(F, d(funcpar(F, FForm, T), Func)) | Env], Store) :-
    makeEnvStore(Form, Ac, Env, Store).

/*
**    Procedural parameters.
*/

makeEnvStore([procpar(P, PForm) | Form], [Proc | Ac],
             [e(P, d(procpar(P, PForm), Proc)) | Env], Store) :-
    makeEnvStore(Form, Ac, Env, Store).

makeEnvStore([], [], [], []).

/*
**    Run-time support for the handling of stores.
**    --------------------------------------------
*/

:- dynamic(currentLocNum/1).

/*
**    getNewLoc(N)
**        - N is an integer.
**
**    Instantiate N to a new location, that is, to an integer
**    representing a location.  The next invocation will yield the
**    successor of that integer and so forth. This characteristics is
**    crucial for the implementation of arrays.  The first location is
**    0: numbering is initialized by the run command (see `commands').
*/

getNewLoc(Num) :-
    retract(currentLocNum(Num1)),
    !,
    Num is Num1+1,
    asserta(currentLocNum(Num)).
getNewLoc(0) :-
    asserta(currentLocNum(0)). /* Unused. */

/*
**    getStore(Store, Loc, Val)
**        - Store is a store ([s(l1, v1), ..., s(lk, vk)]);
**        - Loc   is a location;
**        - Val   is a constant (tt, ff or an integer).
**
**    True if, in the store Store, location Loc is associated
**    to the storable value Val.
*/

getStore([s(L, V)|_], L, V) :- !.
getStore([_|T], L, V) :-
    getStore(T, L, V).

/*
**    changeStore(Store, Loc, NewVal, NewStore)
**        - Store, NewStore are stores;
**        - Loc             is a location;
**        - NewVal          is a storable value (tt, ff or an integer).
**
**    True if NewStore is obtained from Store by changing the asssociation
**    <Loc, Store(Loc)> to <Loc, Val>.
*/

changeStore(S, Loc, Val, [s(Loc, Val)|S1]) :-
    clair_delete(s(Loc, _), S, S1).

/*
**    dispose(Env, Store, NewStore) -
**        - Env             is a value environment;
**        - Store, NewStore are stores.
**
**    This predicate is used at the end of the evaluation of the
**    command blocks "d ; c".
**    True if NewStore is obtained from store by:
**
**        1) Removing all the associations <Loc, SVal> such that
**           <_, l(Loc)> is an element of Env.  This means that all
**           the locations corresponding to variables declared in
**           the block are disposed.
**
**        2) Removing all the associations <Loc, SVal> that belong
**           to arrays in the codomain of Env.  This means that all
**           such array locations are disposed.
**
**        3) Copying Store(L2) into Store(L1) and removing the
**           association <L2, SVal> for each parameter passed by
**           value-result, c(L1, L2), that occurs in Env.  This is the
**           final phase of parameter passing by value-result.
**           FIXME: this also means that `dispose' is an exceptionally
**                  bad name for this operation.
*/

dispose([e(_, d(_, l(L))) | T], S, S1) :-
    clair_delete(s(L, _), S, S2),
    dispose(T, S2, S1).

dispose([e(_, d(_, a(L, B))) | T], S, S1) :-
    numElem(L, N),
    disposeArray(B, N, S, S2),
    dispose(T, S2, S1).

dispose([e(_, d(_, c(L1, L2))) | T], S, S1) :-
    getStore(S, L2, C),
    clair_delete(s(L2, _), S, S2),
    changeStore(S2, L1, C, S3),
    dispose(T, S3, S1).

dispose([_ | T], S, S1) :-
    dispose(T, S, S1).

dispose([], S, S).

disposeArray(_, 0, S, S).
disposeArray(B, N, S, S1) :-
    clair_delete(s(B, _), S, S2),
    C is B+1,
    M is N-1,
    disposeArray(C, M, S2, S1).

/*
**    clair_delete(Elem, List, NewList)
**        - Elem          is a term;
**        - List, NewList are lists.
**
**    True if NewList is obtained from List by removing the first
**    element that unifies with Elem.
*/

clair_delete(X, [X|Y], Y).
clair_delete(X, [Y|L], [Y|NL]) :-
    clair_delete(X, L, NL).

/*
**    readInteger(X)
**
**    Read an integer from the current input stream.
*/

readInteger(Res) :-
    get_code(Char), ((Char = 45) ->
        (
	    get_code(Char1),
	    readInteger_2(Char1, Res1),
	    Res is -Res1)
	;
	    readInteger_2(Char, Res)
	).

readInteger_2(Char, Result) :-
    digit(Char, _),
    !,
    readInteger_3(Char, 0, Result).
readInteger_2(Char, Result) :-
    readString_1(Char, _),    /* Ignore until end-of-line. */
    write('Sorry, an integer is expected.'), nl,
    readInteger(Result).

readInteger_3(Digit, Accu, Result) :-
    digit(Digit, Digit_weight),
    !,
    get_code(Inter_char), /* Overtaking `Digit'. */
    Inter_accu is Accu*10+Digit_weight,
    readInteger_3(Inter_char, Inter_accu, Result).
readInteger_3(_, Accu, Accu).

digit(X, Weight) :-
    X >= 48,
    X =< 57,
    Weight is X-48.

/*
**    readString(S)
**
**    Read a string (until end-of-line) from the current input stream.
*/

readString(String) :-
    get_code(Char),
    readString_1(Char, String).

/* ascii(<CR>, 10) */

readString_1(10, []) :-
    ! .
readString_1(Char, [Char|Tail]) :-
    get_code(Next),
    readString_1(Next, Tail).


/*
**    readAtom(A)
**
**    Read a Prolog atom from the current input stream.
*/

readAtom(Atom) :-
    get_code(Char),
    readAtom_1(Char, String),
    returnAtom(String, Atom).

readAtom_1(Char, [Char|Tail]) :-
    atomChar(Char),
    !,
    get_code(Next), readAtom_1(Next, Tail).
readAtom_1(_, []) :- ! .

atomChar(C) :-
    C >= 97, C =< 122 .  /* lower-case letters */
atomChar(C) :-
    C >= 64, C =< 90 .   /* upper-case letters */
atomChar(C) :-
    C >= 47, C =< 57 .   /* digits */
atomChar(39). /* ascii(39,'''') */
atomChar(45). /* ascii(45,'-') */
atomChar(95). /* ascii(95,'_') */

returnAtom([], []) :-
    ! .
returnAtom(String, Atom) :-
    name(Atom, String).
