/* Predicates for the static semantics checks and functions.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `statsem'
**
**    This module contains the predicates implementing the static
**    semantics checks for SIL and SFL.
**
**    Predicates defined in this module:
**
**    aeMatchForm/3, cFV/2, dDV/2, dFV/2, dssem/2, eFV/2, elFV/2,
**    elssem/3, envDomain/2, envRangeFV/2, essem/3, fEnv/2, formDV/2,
**    formMatchForm/2, limitsOk/1, nIndexOk/2, tIndexOk/1, wfc/2, wfd/2.
*/

/*
**                          IMPORTANT
**                          =========
**
**    The predicates wfd/3, essem/3, aeMatchForm/3 and wfc/2 always
**    succeed to allow for the complete analysis of the syntax tree.
**    This increases the chances of generating meaningful and non
**    redundant error messages.
**
**    When an error is detected, the above mentioned predicates invoke
**    error/1, which, besides delivering the appropriate error message
**    to the user, asserts the clause `scheck(failed)'. The presence
**    of this clause allows, at the end of the static semantics
**    analysis, to know whether the program is correct (with respect
**    to types) or not.
*/

/*
**    Clauses for the static semantics of declarations.
**    =================================================
**
**    The following predicates will now be defined:
**
**    dFV(D, V)
**       - D is a declaration;
**       - V is a set of variables (represented by a list).
**
**    True if V is the set of free variables in D.
**
**    dDV(D, V)
**       - D is a declaration;
**       - V is a set of variables (represented by a list).
**
**    True if V is the set of definition variables of  D.
**
**    wfd(A, D)
**       - A is a type environment;
**       - D is a declaration.
**
**    Always true.
**    If D is not well formed in A, the clause `scheck(failed)' is
**    asserted and an appropriate error message is given (see above).
**
**    dssem(D, A)
**       - D is a declaration;
**       - A is a type environment.
**
**    True if A is the type environment generated by declaration D.
*/

:- discontiguous(dFV/2).
:- discontiguous(dDV/2).
:- discontiguous(wfd/2).
:- discontiguous(dssem/2).

/*
**    Nil.
**    ----
*/

dFV(nil, []).
dDV(nil, []).
wfd(_, nil).
dssem(nil, []).

/*
**   Simple declaration (SFL only).
**   ------------------------------
*/

dFV(def(_, _, E), V) :-
    eFV(E, V).
dDV(def(X, _, _), [X]).
wfd(A, def(X, T, E)) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, varDecl, X, T, Te])).
dssem(def(X, T, _), [e(X, T)]).

/*
**    Constant declaration (SIL only).
**    --------------------------------
*/

dFV(con(_, _, E), V) :-
    eFV(E, V).
dDV(con(X, _, _), [X]).
wfd(A, con(X, T, E)) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, constDecl, X, T, Te])).
dssem(con(X, T, _), [e(X, T)]).

/*
**    Scalar and array variable declaration (SIL only).
**    -------------------------------------------------
*/

dFV(var(_, _, E), V) :-
    eFV(E, V).
dDV(var(X, _, _), [X]).
wfd(A, var(X, array(Limits, T), E)) :-
    !,
    (limitsOk(Limits) ->
        true
    ;
        error([malformed, arrayDecl, X])),
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, arrayDecl, X, T, Te])).

wfd(A, var(X, T, E)) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, varDecl, X, T, Te])).

dssem(var(X, array(Limits, T), _), [e(X, a(Limits, T))]) :- !.
dssem(var(X, T, _), [e(X, l(T))]).

/*
**    limitsOk(Limits)
**        - Limits is a list of array limits.
**
**    True if, for each limit <L, H> in Limits, we have L <= H.
*/

limitsOk([b(L, H)|T]) :-
    L =< H,
    limitsOk(T).
limitsOk([]).

/*
**    Procedure declaration (SIL only).
**    ---------------------------------
*/

dFV(procdef(_, Formals, C), V) :- 
    cFV(C, V1), formDV(Formals, V2), 
    subtraction(V1, V2, V).
dDV(procdef(P, _, _), [P]).
wfd(A, procdef(P, Formals, C)) :-
    formDV(Formals, V),
    moreThanOnce(V, Duplicates),
    (Duplicates=[] ->
        true
    ;
        error([malformed, procDecl, P, Duplicates])),
    fEnv(Formals, Fenv),
    compose(A, Fenv, G),
    wfc(G, C).

dssem(procdef(P, Formals, _), [e(P, ptype(Formals))]).

/*
**    Sequential composition.
**    -----------------------
*/

dFV(seq(D1, D2), V) :- 
    dFV(D1, FVd1), dDV(D1, DVd1), dFV(D2, FVd2), 
    subtraction(FVd2, DVd1, RFVd2), 
    clair_union(FVd1, RFVd2, V).
dDV(seq(D1, D2), V) :- 
    dDV(D1, DVd1), dDV(D2, DVd2), 
    clair_union(DVd1, DVd2, V).
wfd(A, seq(D1, D2)) :- 
    wfd(A, D1),
    dssem(D1, B), 
    compose(A, B, G),
    wfd(G, D2).
dssem(seq(D1, D2), G) :- 
    dssem(D1, B1),
    dssem(D2, B2),
    compose(B1, B2, G).

/*
**    Parallel composition.
**    ---------------------
*/

dFV(parallel(D1, D2), V) :-
    dFV(D1, FVd1), dFV(D2, FVd2),
    clair_union(FVd1, FVd2, V).
dDV(parallel(D1, D2), V) :-
    dDV(D1, DVd1), dDV(D2, DVd2),
    clair_union(DVd1, DVd2, V).
wfd(A, parallel(D1, D2)) :-
    wfd(A, D1),
    wfd(A, D2),
    dDV(D1, V1), dDV(D2, V2),
    clair_intersection(V1, V2, V),
    (V=[] ->
        true
    ;
        error([malformed, parallel, V])).

dssem(parallel(D1, D2), G) :-
    dssem(D1, B1),
    dssem(D2, B2),
    compose(B1, B2, G).

/*
**    Private composition.
**    --------------------
*/

dFV(private(D1, D2), V) :- 
    dFV(D1, FVd1), dDV(D1, DVd1), dFV(D2, FVd2),
    subtraction(FVd2, DVd1, RFVd2),
    clair_union(FVd1, RFVd2, V).
dDV(private(_, D2), V) :- 
    dDV(D2, V).
wfd(A, private(D1, D2)) :-
    wfd(A, D1),
    dssem(D1, B),
    compose(A, B, G),
    wfd(G, D2).
dssem(private(_, D2), G) :- 
    dssem(D2, G).

/*
**    Function declaration.
**    ---------------------
*/

dFV(fundef(_, Formals, _, E), V) :- 
    eFV(E, V1), formDV(Formals, V2), 
    subtraction(V1, V2, V).
dDV(fundef(F, _, _, _), [F]).
wfd(A, fundef(F, Formals, T, E)) :-
    formDV(Formals, V),
    moreThanOnce(V, Duplicates),
    (Duplicates=[] ->
        true
    ;
        error([malformed, funDecl, F, Duplicates])),
    fEnv(Formals, Fenv),
    compose(A, Fenv, G),
    essem(G, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, funDecl, F, T, Te])).

dssem(fundef(F, Formals, T, _), [e(F, ftype(Formals, T))]).

/*
**    Recursive declaration.
**    ----------------------
*/

dFV(rec(D), V) :- 
    dFV(D, FVd), dDV(D, DVd),
    subtraction(FVd, DVd, V).
dDV(rec(D), V) :- 
    dDV(D, V).
wfd(A, rec(D)) :- 
    dssem(D, B),
    dFV(D, FVd), dDV(D, DVd),
    clair_intersection(FVd, DVd, V0),
    projection(B, V0, B0), 
    compose(A, B0, G),
    wfd(G, D).

dssem(rec(D), A) :- 
    dssem(D, A).

/*
**    Parameter passing (only for the extended abstract syntax).
**    ----------------------------------------------------------
*/

dFV(parPass(_, Ae), V) :-
    elFV(Ae, V).
dDV(parPass(Form, _), V) :-
    formDV(Form, V).

/*
**    Notice that we do not define wfd/2 and dssem/2 for terms of the
**    form parPass(Form, Ae) since they never appear in initial
**    configurations.
*/

/*
**    Value environment (in the extended abstract syntax it is a declaration).
**    ------------------------------------------------------------------------
*/

dFV(E, V) :- 
    isEnv(E),
    envRangeFV(E, V).
dDV(E, V) :- 
    isEnv(E),
    envDomain(E, V).

/*
**    Notice that we do not define wfd/2 and dssem/2 for environments
**    since they never appear in initial configurations.
*/

/*
**    envRangeFV(E, V)
**        - E is a value environment;
**        - V is a set of variables.
**
**    True if V is the set of free variables in the codomain of E.
**    (This is required for recursive declarations.)
*/

envRangeFV([e(_, func(Formals, Body))|E], V) :-
    eFV(Body, FVbody), formDV(Formals, DVformals),
    subtraction(FVbody, DVformals, V1),
    envRangeFV(E, V2), clair_union(V1, V2, V).

envRangeFV([e(_, proc(Formals, Body))|E], V) :-
    cFV(Body, FVbody), formDV(Formals, DVformals),
    subtraction(FVbody, DVformals, V1),
    envRangeFV(E, V2), clair_union(V1, V2, V).

envRangeFV([_|E], V) :-
    envRangeFV(E, V).
envRangeFV([], []).

/*
**    envDomain(E, V)
**        - E is a value environment;
**        - V is a set of variables.
**
**    True if V is the domain of (the set of variables defined in) E.
*/

envDomain([e(X, _)|E], V) :-
    envDomain(E, V0), clair_union([X], V0, V).
envDomain([], []).

/*
**    Formal parameters.
**    ------------------
*/

/*
**    formDV(FormList, V)
**        - FormList is a list of formal parameters;
**        - V is a set of variables (represented by a list).
**
**    True if V is the set of definition variables of FormList.
*/

formDV([   ival(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([   fval(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([   name(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([    ref(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([  const(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([   copy(X, _)    | Form], [X | V]) :- formDV(Form, V).
formDV([funcpar(N, _, _) | Form], [N | V]) :- formDV(Form, V).
formDV([procpar(N, _)    | Form], [N | V]) :- formDV(Form, V).
formDV([], []).

/*
**    fEnv(FormList, Env)
**        - FormList is a list of formals;
**        - Env is a type environment.
**
**    True if Env is the type environment generated by FormList.
*/

fEnv([      ival(X, T) | Form], [e(X, l(T))        | A]) :- fEnv(Form, A).
fEnv([      fval(X, T) | Form], [e(X, T)           | A]) :- fEnv(Form, A).
fEnv([      name(X, T) | Form], [e(X, T)           | A]) :- fEnv(Form, A).
fEnv([       ref(X, T) | Form], [e(X, l(T))        | A]) :- fEnv(Form, A).
fEnv([     const(X, T) | Form], [e(X, T)           | A]) :- fEnv(Form, A).
fEnv([      copy(X, T) | Form], [e(X, l(T))        | A]) :- fEnv(Form, A).
fEnv([funcpar(N, F, T) | Form], [e(N, ftype(F, T)) | A]) :- fEnv(Form, A).
fEnv([   procpar(N, F) | Form], [e(N, ptype(F))    | A]) :- fEnv(Form, A).
fEnv([], []).

/*
**    Clauses for the static semantics of expressions.
**    ================================================
**
**    The following predicates will now be defined:
**
**    eFV(E, V)
**        - E is an expression;
**        - V is a set of variables (represented by a list).
**
**    True if V is the set of free variables in E.
**
**    essem(A, E, T)
**      - A is a type environment;
**      - E is an expression;
**      - T is a type.
**
**    Always true.
**    If, in A, E is not well formed or has not type T, the clause
**    `scheck(failed)' is asserted and an appropriate error message is
**    given (see above).
*/

:- discontiguous(eFV/2).
:- discontiguous(essem/3).

/*
**    Boolean constants.
**    ------------------
*/

eFV(tt, []).
essem(_, tt, bool).

eFV(ff, []).
essem(_, ff, bool).

/*
**    Identifiers.
**    ------------
*/

eFV(i(X), [X]).
essem(A, i(X), T) :-
    (getEnv(A, X, W) ->
        ((W=l(T); W=T) ->
            true
        ; 
            error([typeMismatch, id, X, T, W]))
    ;
        error([undeclared, id, X])).

/*
**    Array reference (SIL only).
**    ---------------------------
*/

eFV(aref(X, El), V) :- 
    elFV(El, FVel),
    clair_union([X], FVel, V).
essem(A, aref(X, El), T) :-
    (getEnv(A, X, W) ->
        (W=a(Limits, T) ->
            true
        ;
            error([typeMismatch, array, X, T, W]))
    ;
        error([undeclared, array, X])),
    elssem(A, El, Tl),
    (tIndexOk(Tl) ->
        true
    ;
        error([typeMismatch, arrayIndexType, X])),
    (nIndexOk(Tl, Limits) ->
        true
    ;
        error([typeMismatch, arrayIndexNumber, X])).

/*
**    nIndexOk(TypeL, Limits)
**        - TypeL  is a list of types;
**        - Limits is a list of array limits.
**
**    True if TypeL and Limits have the same length. This is used to
**    check that array references have the correct number of indices.
*/

nIndexOk([_|Tl], [_|Limits]) :- 
    nIndexOk(Tl, Limits).
nIndexOk([], []).

/*
**    tIndexOk(TypeL)
**        - TypeL is a list of types.
**
**    True if all the elements of TypeL are `int' (the abstract syntax
**    representation of the integer type). This is used to check that
**    the expressions for the indices in an array references have the
**    right type.
*/

tIndexOk([int|Tl]) :- 
    tIndexOk(Tl).
tIndexOk([]).

/*
**    Boolean negation.
**    -----------------
*/

eFV(not(X), V) :-
    eFV(X, V).
essem(A, not(X), bool) :-
    essem(A, X, T),
    (T=bool ->
        true
    ;
        error([typeMismatch, uop, not, bool, T])).

/*
**    Unary minus.
**    ------------
*/

eFV(uminus(X), V) :-
    eFV(X, V).
essem(A, uminus(X), int) :-
    essem(A, X, T),
    (T=int ->
        true
    ;
        error([typeMismatch, uop, uminus, int, T])).

/*
**    Expression block.
**    -----------------
*/

eFV(let(D, E), V) :-
    eFV(E, FVe), dDV(D, DVd),
    subtraction(FVe, DVd, V).
essem(A, let(D, E), T) :-
    wfd(A, D),
    dssem(D, B),
    compose(A, B, G),
    essem(G, E, T).

/*
**    Closure block for expression abstractions.
**    ------------------------------------------
*/

/* It has no free variables by definition. */
/* FIXME: what is this clause here for? */
eFV(closureE(_, _), []) :-
    write('!!! eFV(closureE) !!!'), nl.

/*
**    Notice that we do not define essem/3 for closure blocks since
**    they never appear in initial configurations.
*/

/*
**    Conditional expression.
**    -----------------------
*/

eFV(if(E, E1, E2), V) :- 
    eFV(E, FVe), eFV(E1, FVe1), eFV(E2, FVe2),
    clair_union(FVe1, FVe2, FVe1e2), 
    clair_union(FVe, FVe1e2, V).
essem(A, if(E, E1, E2), T) :-
    essem(A, E, T0),
    (T0=bool ->
        true
    ;
        error([typeMismatch, if, T0 ])),
    essem(A, E1, T1), essem(A, E2, T2), T=T1,
    (T1=T2 ->
        true
    ;
        error([typeMismatch, if, T1, T2])).

/*
**    Function call.
**    --------------
*/

eFV(funcall(F, Ae), V) :-
    elFV(Ae, FVae),
    clair_union([F], FVae, V).

essem(A, funcall(F, Ae), T) :-
    (getEnv(A, F, ftype(Formals, T)) ->
        (aeMatchForm(A, Ae, Formals) ->
            true
        ;
            error([typeMismatch, function, wrongNumberOfActuals, F]))
    ;
        error([undeclared, function, F])).

/*
**    Input.
**    ------
*/

eFV(input(_), []).
essem(_, input(T), T).

/* 
**    Expressions with side-effects (SIL only).
**    -----------------------------------------
*/

eFV(side(D, C, E), V) :-
    dDV(D, DVd), dFV(D, FVd),
    cFV(C, FVc), eFV(E, FVe),
    clair_union(FVc, FVe, V1),
    subtraction(V1, DVd, V2),
    clair_union(V2, FVd, V).
essem(A, side(D, C, E), T) :-
    wfd(A, D),
    dssem(D, B),
    compose(A, B, G),
    wfc(G, C),
    essem(G, E, T).

/*
**    Integer constants.
**    ------------------
*/

eFV(X, []) :- 
    integer(X).
essem(_, X, int) :-
    integer(X).

/*
**    Binary operators.
**    -----------------
*/

eFV(X, V) :-
    X =.. [O, Arg1, Arg2],
    tbop(O, _, _, _, _),
    eFV(Arg1, V1), eFV(Arg2, V2),
    clair_union(V1, V2, V).
essem(A, X, T) :-
    X =.. [O, Arg1, Arg2],
    tbop(O, T1, T2, T, _),
    essem(A, Arg1, T1a),
    essem(A, Arg2, T2a),
    ((T1=T1a, T2=T2a) ->
        true
    ;
        error([typeMismatch, bop, O, T1a, T2a])).

/*
**    Lists of expressions.
**    ---------------------
*/

/*
**    elFV(E, V)
**      - E is a list of expressions
**      - V is the set of free variables in E.
*/

/*
**    elssem(A, El, Tl) 
**      - A is a type environment,
**      - El is a list of expressions,
**      - Tl is the corresponding list of types.
*/

elFV([E | Ae], V) :- 
    eFV(E, FVe), elFV(Ae, FVae),
    clair_union(FVe, FVae, V).
elFV([], []).
elssem(A, [E | Ae], [T | Aet]) :- 
    essem(A, E, T),
    elssem(A, Ae, Aet).
elssem(_, [], []).

/*
**    Clauses for the static semantics of commands.
**    =============================================
**
**    The following predicates will now be defined:
**    
**    cFV(C, V)
**        - C is a command;
**        - V is a set of variables (represented by a list).
**
**    True if V is the set of free variables in command C.
**
**    wfc(A, C)
**        - A is a type environment;
**        - C is a command.
**
**    Always true.
**    If C is not well formed in A, the clause `scheck(failed)' is
**    asserted and an appropriate error message is given (see above).
*/

:- discontiguous(cFV/2).
:- discontiguous(wfc/2).

/*
**    Nop.
**    ----
*/

cFV(nop, []).
wfc(_, nop).

/*
**    Assignment.
**    -----------
*/

cFV(ass(X, E), V) :-
    eFV(X, FVx), eFV(E, FVe),
    clair_union(FVx, FVe, V).
wfc(A, ass(aref(X, El), E)) :-
    essem(A, aref(X, El), T),
    essem(A, E, T).
wfc(A, ass(i(X), E)) :-
    essem(A, E, T1),
   (getEnv(A, X, Y) ->
       (Y=l(T) ->
           (T=T1 ->
               true
           ;
               error([typeMismatch, ass, X, T, T1]))
       ; 
           error([nolvalue, ass, X]))
   ;
       error([undeclared, var, X])).

/*
**    Command sequence.
**    -----------------
*/

cFV(seq(C1, C2), V) :-
    cFV(C1, V1), cFV(C2, V2),
    clair_union(V1, V2, V).
wfc(A, seq(C1, C2)) :-
    wfc(A, C1),
    wfc(A, C2).

/*
**    Conditional command.
**    --------------------
*/

cFV(if(E, C1, C2), V) :-
    eFV(E, FVe),
    cFV(C1, V1), cFV(C2, V2),
    clair_union(V1, V2, V12),
    clair_union(FVe, V12, V).
wfc(A, if(E, C1, C2)) :-
    essem(A, E, T),
    (T=bool ->
        true
    ;
        error([typeMismatch, cif, T])),
    wfc(A, C1), wfc(A, C2).

/*
**    While command.
**    --------------
*/

cFV(while(E, C), V) :-
    eFV(E, FVe), cFV(C, FVc),
    clair_union(FVe, FVc, V).
wfc(A, while(E, C)) :-
    essem(A, E, T),
    (T=bool ->
        true
    ;
        error([typeMismatch, while, T])),
    wfc(A, C).

/*
**    Command block.
**    --------------
*/

cFV(block(D, C), V) :-
    dDV(D, DVd), dFV(D, FVd), cFV(C, FVc),
    subtraction(FVc, DVd, V0),
    clair_union(FVd, V0, V).
wfc(A, block(D, C)) :-
    wfd(A, D),
    dssem(D, B),
    compose(A, B, G),
    wfc(G, C).

/*
**    Closure block for procedural abstractions.
**    ------------------------------------------
*/

/* It has no free variables by definition. */
/* FIXME: what is this clause here for? */
cFV(closureC(_, _), []) :-
    write('!!! cFV(clocureC) !!!'), nl.

/*
**    Notice that we do not define wfc/3 for closure blocks since
**    they never appear in initial configurations.
*/

/*
**    Procedure call.
**    ---------------
*/

cFV(proccall(P, Ae), V) :-
    elFV(Ae, FVae),
    clair_union([P], FVae, V).
wfc(A, proccall(P, Ae)) :-
    (getEnv(A, P, ptype(Formals)) ->
        (aeMatchForm(A, Ae, Formals) ->
            true
        ;
            error([typeMismatch, procedure, wrongNumberOfActuals, P]))
    ;
        error([undeclared, procedure, P])).

/*
**    Print command.
**    --------------
*/

cFV(print(_, El), V) :-
    elFV(El, V).
wfc(A, print(_, El)) :-
    elssem(A, El, _).

/*
**    aeMatchForm(A, Ae, Form)
**        - A    is a type environment;
**        - Ae   is a list of actual parameters;
**        - Form is a list of formal parameters.
**
**    Always true.
**    If, in the type environment A, the actual parameters Ae are not
**    consistent with the formal parameters Form, the clause
**    `scheck(failed)' is asserted and an appropriate error message is
**    given (see above).
**
**    N.B. Consistency of an actual parameter with a formal parameter:
**
**         1) parameter passing by value, name and constant:
**              a) the types must coincide;
**         2) parameter passing by reference and value-result:
**              a) the types must coincide;
**              b) the actual must be an lvalue (variable or array);
**         3) functional parameters:
**              a) the actual parameters must be functions (obviously);
**              b) the formals of the parameters must coincide in number,
**                 type and parameter passing mechanism;
**         4) procedural parameters:
**              a) the actual parameters must be procedures (obviously);
**              b) the formals of the parameters must coincide in number,
**                 type and parameter passing mechanism.
*/

/*
**    Call by value (SIL only).
**    -------------------------
*/

aeMatchForm(A, [E | El], [ival(X, T) | Form]) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    aeMatchForm(A, El, Form).

/*
**    Call by value (SFL only).
**    -------------------------
*/

aeMatchForm(A, [E | El], [fval(X, T) | Form]) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    aeMatchForm(A, El, Form).

/*
**    Call by name.
**    -------------
*/

aeMatchForm(A, [E | El], [name(X, T) | Form]) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    aeMatchForm(A, El, Form).

/*
**    Call by reference (SIL only).
**    -----------------------------
*/

aeMatchForm(A, [E | El], [ref(X, T) | Form]) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    (((E=i(Y), getEnv(A, Y, l(_))) ; E=aref(_, _)) ->
        true
    ;
        error([nolvalue, callByReference, X])),
    aeMatchForm(A, El, Form).

/*
**    Call by constant (SIL only).
**    ----------------------------
*/

aeMatchForm(A, [E | El], [const(X, T) | Form]) :- 
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    aeMatchForm(A, El, Form).

/*
**    Call by value-result (SIL only).
**    --------------------------------
*/

aeMatchForm(A, [E | El], [copy(X, T) | Form]) :-
    essem(A, E, Te),
    (T=Te ->
        true
    ;
        error([typeMismatch, formal, X, T, Te])),
    (((E=i(Y), getEnv(A, Y, l(_))) ; E=aref(_, _)) ->
        true
    ;
        error([nolvalue, callByValueResult, X])),
    aeMatchForm(A, El, Form).

/*
**    Functional parameters.
**    ----------------------
*/

aeMatchForm(A, [E | El], [funcpar(N, F, T) | Form]) :-
    (E=i(X) ->
        (getEnv(A, X, De) ->
            (De=ftype(FormE, Te) ->
                ((formMatchForm(F, FormE) ->
                    true
                ;
                    error([formalMismatch, function, N, F, X, FormE]))),
                (T=Te ->
                    true
                ;
                    error([typeMismatch, functionReturn, N, T, X, Te]))
            ;
                error([typeMismatch, notAFunction, id, N, X, De]))
        ;
            error([undeclared, function, X]))
    ;
        error([typeMismatch, notAFunction, expr, N])),
    aeMatchForm(A, El, Form).

/*
**    Procedural parameters (SIL only).
**    ---------------------------------
*/

aeMatchForm(A, [E | El], [procpar(N, F) | Form]) :-
    (E=i(X) ->
        (getEnv(A, X, De) ->
            (De=ptype(FormE) ->
                (formMatchForm(F, FormE) ->
                    true
                ;
                    error([formalMismatch, procedure, N, F, X, FormE]))
            ;
                error([typeMismatch, notAProcedure, id, N, X, De]))
        ;
            error([undeclared, procedure, X]))
    ;
        error([typeMismatch, notAProcedure, expr, N])),
    aeMatchForm(A, El, Form).

aeMatchForm(_, [], []).

/*
**    formMatchForm(F1, F2)
**        - F1, F2 are lists of formal parameters.
**
**    True if F1 and F2 coincide in the number, order, type and
**    parameter passing mechanism of the formals they contain.
*/

formMatchForm([   ival(_, T)    | T1], [   ival(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([   fval(_, T)    | T1], [   fval(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([   name(_, T)    | T1], [   name(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([    ref(_, T)    | T1], [    ref(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([  const(_, T)    | T1], [  const(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([   copy(_, T)    | T1], [   copy(_, T)    | T2]) :-
    formMatchForm(T1, T2).
formMatchForm([funcpar(_, F1, T) | T1], [funcpar(_, F2, T) | T2]) :-
    formMatchForm(F1, F2), formMatchForm(T1, T2).
formMatchForm([procpar(_, F1)    | T1], [procpar(_, F2)    | T2]) :-
    formMatchForm(F1, F2), formMatchForm(T1, T2).
formMatchForm([], []).
