/* Implementation of the tracing facilities.
   Copyright (C) 2002-2005 Roberto Bagnara <bagnara@cs.unipr.it>

This file is part of the Combined Language and Abstract Interpretation
Resource (CLAIR).

CLAIR is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2 of the License, or (at your
option) any later version.

CLAIR is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
USA.

For the most up-to-date information see the CLAIR site:
http://www.cs.unipr.it/clair/ . */

/*
**    CLAIR: module `trace'
**
**    This module contains the predicates implementing pretty-printing
**    of the abstract syntax tree of the program, as generated bby the
**    parser and then modified by the dynamic semantics predicates.
**
**    Predicates defined in this module:
**
**    nextStep/1, outputConfig/1, outputStore/1, pS/2, pST/2, pStore/1,
**    pp/1, ppAct/2, ppStore/1, ppae/2, ppc/3, ppd/3, ppe/3, ppel/2,
**    ppenv/1, ppenvstuff/1, ppf/1, ppfs/1, ppin/1, ppop/1, ppt/1,
**    s/1, tabS/2. 
*/

/*
**    outputConfig(X)
**
**    Append the pretty-print of configuration X the the current
**    trace file.
*/

outputConfig(_) :-
    traceConfig(no), !.
outputConfig(X) :-
    traceFile(F),
    open(F, append, Stream),
    current_output(OldStream),
    set_output(Stream),
    nextStep(N),
    (N=\=0 ->
        (write('=============================== step '),
        write(N), write(' ==============================='), nl)
    ;
        true),
    pp(X), !,
    close(Stream),
    set_output(OldStream).

/*
**    outputStore(S)
**
**    Append the pretty-print of store S to the current trace file.
*/

outputStore(_) :-
    traceStore(no), !.
outputStore([]) :- !.
outputStore(X) :-
    traceFile(F),
    open(F, append, Stream),
    current_output(OldStream),
    set_output(Stream),
    write('-----------------------------------'),
    write('-----------------------------------'), nl,
    ppStore(X), !,
    close(Stream),
    set_output(OldStream).

/*
**    pp(X)
**
**    Print the configuration encoded by X.
*/

pp(X) :-
    (language(functional) ->
        ppe(0, 0, X)
    ;
        ppc(0, 0, X)),
    nl.

/*
**    s(N)
**
**    Tabulator: does nothing if N=0; otherwise starts a new line with
**    N blanks on the current trace file.
*/

s(0) :- !.
s(N) :- nl, tab(N).

/*
**    ppt(Type)
**
**    Print the type Type.
*/

ppt(int) :-
     write('integer').
ppt(bool) :-
     write('boolean').
ppt(array(L, T)) :-
     write('array ['),
     ppin(L),
     write('] of '),
     ppt(T).

/*
**    ppin(Limits)
**
**    Print the limits Limits of an array.
*/

ppin([b(B, L)]) :-
    write(B), write('..'), write(L).
ppin([H|T]) :-
    ppin([H]),
    write(','),
    ppin(T).
ppin([]).

/*
**    ppel(El)
**
**    Print a list El of expressions.
*/
 
ppel(W, [E]) :-
    ppe(W, 0, E).
ppel(W, [E|El]) :-
    ppe(W, 0, E),
    write(','),
    ppel(W, El).
ppel(_, []).

/*
**    Pretty printing of expressions.
*/

/*
**    ppe(W, N, E)
**      - W is the current indentation;
**      - N is 0 for normal printing;
**          otherwise printing must be preceded by a newline and N blanks;
**      - E is the expression to be printed.
*/

/*
**    Boolean constants.
*/

ppe(_, N, tt) :-
    s(N),
    write(true).
ppe(_, N, ff) :-
    s(N),
    write(false).

/*
**    Identifiers.
*/

ppe(_, N, i(X)) :-
    s(N),
    write(X).
ppe(W, N, aref(X, El)) :- 
    s(N),
    write(X), write('['),
    ppel(W, El),
    write(']').

/*
**    Boolean negation.
*/

ppe(W, N, not(X)) :-
    s(N),
    write('not '),
    M is W+2, ppe(M, 0, X).

/*
**    Unary minus.
*/

ppe(W, N, uminus(X)) :-
    s(N),
    write('-'),
    M is W+2, ppe(M, 0, X).

/*
**    Expression block.
*/

ppe(W, _, let(D, E)) :-
    s(W),
    write('let'),
    M is W+2, ppd(M, M, D), 
    nl, tab(W), write('in'),
    ppe(M, M, E).

/*
**    Blocco per la chiusura delle astrazioni delle espressioni:
**    funzioni e parametri passati per nome.
*/

ppe(W, _, closureE(D, E)) :-
    s(W),
    write('let /* closure, no dispose */'),
    M is W+2, ppd(M, M, D), 
    nl, tab(W), write('in'),
    ppe(M, M, E).

/*
**    Conditional expression.
*/

ppe(W, _, if(E, E1, E2)) :-
    s(W),
    write('(if '),
    M is W+2, ppe(M, 0, E),
    write(' then'),
    ppe(M, M, E1),
    nl, tab(W), write('else'),
    ppe(M, M, E2),
    write(')').

/*
**    Function call.
*/

ppe(W, N, funcall(F, El)) :-
    s(N),
    write(F), write('('),
    M is W+2, ppel(M, El),
    write(')').

/*
**    Input.
*/

ppe(_, N, input(T)) :-
    s(N),
    write('input '),
    ppt(T).

/*
**    Expressions with side-effects.
*/

ppe(W, _, side(D, C, E)) :-
    s(W),
    write('expr'),
    M is W+2, ppd(M, M, D),
    write(';'), 
    ppc(M, M, C),
    nl, tab(W), write('result '),
    ppe(M, M, E).

/*
**    Binary operators.
*/

ppe(W, N, X) :-
    X =.. [O, Arg1, Arg2],
    tbop(O, _, _, _, _),
    (Arg1=if(_, _, _) ; Arg1=let(_, _) ; Arg1=side(_, _, _) ; s(N)),
    M is W+2, ppe(M, 0, Arg1),
    ppop(O),
    ppe(M, 0, Arg2).

/*
**    Integer constant.
*/

ppe(_, N, X) :-
    s(N),
    write(X).

/*
**    ppop(Op)
**
**    Print the operator Op.
*/

ppop(plus) :- write('+').
ppop(minus) :- write('-').
ppop(prod) :- write('*').
ppop(or) :- write(' or ').
ppop(div) :- write(' div ').
ppop(mod) :- write(' mod ').
ppop(and) :- write(' and ').
ppop(eq) :- write('==').
ppop(ne) :- write('<>').
ppop(lt) :- write('<').
ppop(le) :- write('<=').
ppop(ge) :- write('>=').
ppop(gt) :- write('>').

/*
**    Pretty printing of declarations.
*/

/*
**    ppd(W, N, D)
**      - W is the current indentation;
**      - N is 0 for normal printing;
**          otherwise printing must be preceded by a newline and N blanks;
**      - D is the declaration to print.
*/

/*
**    Nil.
*/

ppd(_, N, nil) :-
    s(N),
    write('nil').

/*
**    Simple declaration.
*/

ppd(W, N, def(X, T, Y)) :-
    s(N),
    write(X), write(:),
    ppt(T),
    write('='),
    M is W+2, ppe(M, 0, Y).

/*
**    Constant declaration.
*/

ppd(W, N, con(X, T, Y)) :-
    s(N),
    write('const '), write(X), write(':'),
    ppt(T),
    M is W+2, write('='), ppe(M, 0, Y).

/*
**    Variable declaration.
*/

ppd(W, N, var(X, T, Y)) :-
    s(N),
    write('var '), write(X), write(':'),
    ppt(T),
    M is W+2, write('='), ppe(M, 0, Y).

/*
**    Sequential composition.
*/

ppd(W, N, seq(D1, D2)) :-
    ppd(W, N, D1),
    write(';'),
    ppd(W, N, D2).

/*
**    Parallel composition.
*/

ppd(W, N, parallel(D1, D2)) :-
    ppd(W, N, D1),
    write(' |'),
    ppd(W, N, D2).

/*
**    Private composition.
*/

ppd(W, N, private(D1, D2)) :-
    ppd(W, N, D1),
    write('->'),
    M is N+2, ppd(M, M, D2).

/*
**    Function declaration.
*/

ppd(W, N, fundef(F, Form, T, E)) :-
    s(N),
    write('function '), write(F), write('('),
    ppfs(Form),
    write('):'),
    ppt(T),
    write(' ='),
    M is W+2, ppe(M, M, E).

/*
**    Procedure declaration.
*/

ppd(W, N, procdef(P, Form, C)) :-
    s(N),
    write('procedure '), write(P), write('('),
    ppfs(Form),
    write(')'),
    M is W+2, ppc(M, M, C).

/*
**    Recursive declaration.
*/

ppd(W, N, rec(D)) :-
    s(N),
    write('rec'),
    M is W+2, ppd(M, M, D).

/*
**    Parameter passing.
*/

ppd(W, N, parPass(Fl, Ae)) :-
    s(N),
    write('<'),
    ppfs(Fl),
    write('>=<'),
    M is W+2, ppae(M, Ae),
    write('>').

/*
**    Environment.
*/

ppd(_, N, E) :-
    s(N),
    isEnv(E),
    write('{'),
    ppenv(E),
    write('}').

ppd(_, N, X) :-
    s(N),
    write(X).

/*
**    ppenv(Env)
**
**    Print the environment Env.
*/

ppenv([e(X, V)]) :-
    write(X), write('='),
    ppenvstuff(V).
ppenv([H|T]) :-
    ppenv([H]),
    write(','),
    ppenv(T).
ppenv([]).

/*
**    ppenvstuff(D)
**
**    Pretty printing of the objects that environments associate
**    to identifiers.
*/

/*
** For the SIL's run-time environments, we only print the denotable
** value, ignoring the denotable type.
*/
ppenvstuff(d(_, V)) :-
    ppenvstuff(V).

ppenvstuff(l(L)) :-               /* location */
    write('L'), write(L).
ppenvstuff(r(L)) :-               /* parametro passato per riferimento */
    write('L'), write(L).
ppenvstuff(c(_, L)) :-            /* parametro passato per value-result */
    write('L'), write(L).
ppenvstuff(func(_, _)) :-         /* funzione */
    write('*F*').
ppenvstuff(proc(_, _)) :-         /* procedura */
    write('*P*').
ppenvstuff(a(_, _)) :-            /* array */
    write('*A*').
ppenvstuff(X) :-                  /* costante */
    isConst(X),
    write(X).
ppenvstuff(_) :-                  /* parametro passato per nome */
    write('*N*').

/*
**    ppae(W, Ae)
**   
**    Pretty printing delle liste di parametri attuali Ae.
*/

ppae(W, [Ae]) :-
    ppAct(W, Ae).
ppae(W, [H | Ae]) :-
    ppae(W, [H]),
    write(','),
    ppae(W, Ae).
ppae(_, []).

/*
**    ppAct(W, E)
** 
**    Pretty printing del parametro attuale E.
*/

ppAct(_, func(_, _)) :-
    write('*F*').
ppAct(_, proc(_, _)) :-
    write('*P*').
ppAct(W, Exp) :-
    ppe(W, 0, Exp).
ppAct(_, C) :-
    isConst(C),
    write(C).

/*
**    ppfs(Form)
**
**    Pretty printing dei formal parameters Form.
*/

ppfs([F]) :-
    ppf(F).
ppfs([F|Fl]) :-
    ppfs([F]),
    write(','),
    ppfs(Fl).
ppfs([]).

ppf(ival(X, T)) :-                          /* parametro per valore  */
    write(X), write(':'),                   /* linguaggio imperativo */
    ppt(T).
ppf(fval(X, T)) :-                          /* parametro per valore  */
    write(X), write(':'),                   /* linguaggio funzionale       */
    ppt(T).
ppf(name(X, T)) :-                          /* parametro per nome          */
    write('name '), write(X), write(':'),
    ppt(T).
ppf(ref(X, T)) :-                           /* parametro per riferimento   */
    write('ref '), write(X), write(':'),
    ppt(T).
ppf(const(X, T)) :-                         /* parametro per costante      */ 
    write('const '), write(X), write(':'),
    ppt(T).
ppf(copy(X, T)) :-                          /* parametro per value-result  */
    write('copy '), write(X), write(':'),
    ppt(T).
ppf(funcpar(N, F, T)) :-                    /* parametro funzionale        */
    write('function '), write(N), write('('),
    ppfs(F),
    write('):'),
    ppt(T).
ppf(procpar(N, F)) :-                       /* parametro procedurale      */
    write('procedure '), write(N), write('('),
    ppfs(F),
    write(')').

/*
**    Pretty printing of commands.
*/

/*
**    ppc(W, N, C)
**      - W is the current indentation;
**      - N is 0 for normal printing;
**          otherwise printing must be preceded by a newline and N blanks;
**      - C is the command to be printed.
*/

/*
**    Nop.
*/

ppc(_, N, nop) :-
    s(N),
    write('nop').

/*
**    Assignment.
*/

ppc(W, N, ass(i(X), E)) :- 
    s(N),
    write(X), write(':='),
    M is W+2, ppe(M, 0, E).

ppc(W, N, ass(aref(X, El), E)) :-
    s(N),
    write(X), write('['),
    ppel(W, El),
    write(']:='),
    M is W+2, ppe(M, 0, E). 

/*
**    Sequential composition.
*/

ppc(W, N, seq(C1, C2)) :-
    ppc(W, N, C1),
    write(';'),
    ppc(W, N, C2).

/*
**    Conditional command.
*/

ppc(W, _, if(E, C1, C2)) :-
    s(W),
    write('if '),
    M is W+2, ppe(M, 0, E),
    write(' then'),
    ppc(M, M, C1),
    nl, tab(W), write('else'),
    ppc(M, M, C2).

/*
**    While.
*/

ppc(W, N, while(E, C)) :- 
    s(N),
    write('while '),
    M is W+2, ppe(M, 0, E),
    write(' do '),
    ppc(M, M, C).

/*
**    Comando di output 'print'.
*/

ppc(_, N, print(F, [])) :- !,
    s(N),
    write('print("'), write(F), write('")').

ppc(W, N, print(F, El)) :-
    s(N),
    write('print("'), write(F), write('", '),
    M is W+2, ppel(M, El),
    write(')').

/*
**    Procedure call.
*/

ppc(W, N, proccall(P, El)) :-
    s(N),
    write(P), write('('), 
    M is W+2, ppel(M, El),
    write(')').

/*
**    Blocco.
*/

ppc(_, N, block(D, C)) :-
    s(N),
    write('begin'),
    M is N+2, ppd(M, M, D),
    write(';'), 
    ppc(M, M, C),
    nl, tab(N), write('end').

/*
**    Blocco per la chiusura delle astrazioni procedurali.
*/

ppc(_, N, closureC(D, C)) :-
    s(N),
    write('begin /* closure */'),
    M is N+2, ppd(M, M, D),
    write(';'), 
    ppc(M, M, C),
    nl, tab(N), write('end /* no dispose */').

/*
**    Halt.
*/

ppc(_, N, halt) :-
    s(N),
    write('halt').

ppc(_, N, X) :-
    s(N),
    write(X).

/*
**    Pretty printing of stores.
*/

ppStore(S) :-
    sort(S, S1),
    pStore(S1).

pStore([s(L1, V1), s(L2, V2), s(L3, V3), s(L4, V4)|Rest]) :- !,
    pST(L1, V1), pST(L2, V2), pST(L3, V3), pS(L4, V4), nl,
    pStore(Rest).

pStore([s(L, V)]) :- !,
    pS(L, V), nl.

pStore([s(L, V)|Rest]) :- !,
    pST(L, V),
    pStore(Rest).

pStore([]).

pST(L, V) :-
    pS(L, V),
    tabS(L, V).

pS(L, V) :-
    write('L'), write(L), write(':'), write(V).

tabS(L, V) :-
    name(L, N1), name(V, N2),
    length(N1, L1), length(N2, L2),
    TabNeeded is 20-(L1+L2+2),
    tab(TabNeeded).

/*
**    nextStep(N)
**
**    Istanzia N al numero del prossimo passo effettuato
**    nell'interpretazione.
**    La numerazione viene fatta ripartire da 0 dal comando run
**    (vedi `commands').
*/

:- dynamic(currentStep/1).

nextStep(Num) :-
    retract(currentStep(Num1)),
    !,
    Num is Num1+1,
    asserta(currentStep(Num)).
nextStep(0) :-
    asserta(currentStep(0)). /* Unused. */
